<?php

namespace Drupal\monarch_entity_yml\Commands;

use Drush\Commands\DrushCommands;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;

/**
 * A Drush commandfile.
 *
 * In addition to this file, you need a drush.services.yml
 * in root of your module, and a composer.json file that provides the name
 * of the services file to use.
 *
 * See these files for an example of injecting Drupal services:
 *   - http://cgit.drupalcode.org/devel/tree/src/Commands/DevelCommands.php
 *   - http://cgit.drupalcode.org/devel/tree/drush.services.yml
 */
class EntityYmlCommands extends DrushCommands {

  /**
   * A reference of fields that shouldn't be exported.
   *
   * @var array
   */
  protected const BANNED_FIELDS = [
    'node' => [
      'path' => TRUE,
      'revision_timestamp' => TRUE,
      'revision_uid' => TRUE,
      'revision_log' => TRUE,
      'revision_default' => TRUE,
      'revision_translation_affected' => TRUE,
      'revision_timestamp' => TRUE,
    ],
  ];

  /**
   * Export entity as configuration.
   *
   * @param string $entity_list
   *   Entity ids.
   *
   * @usage entity_yml:dump entity_type:id > file.yml
   *   Export an entity to yml.
   * @usage entity_yml:dump "entity_type:id another_entity_type:*" > file.yml
   *   Export multiple entities to yml.
   *
   * @command entity_yml:dump
   * @aliases eyd
   */
  public function entityExport(string $entity_list = NULL) {
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $data = [];
    $references = [];
    $entities = [];

    preg_match_all('/([^\s,;:]+):([^\s,;:]+)/i', $entity_list, $matches, PREG_SET_ORDER);
    foreach ($matches as $match) {
      try {
        $storage = $entityTypeManager->getStorage($match[1]);

        if (empty($storage)) {
          $this->output()->writeln('Entity storage not found: ' . $match[1]);
          continue;
        }

        if ($match[2] === '*') {
          $res = $storage->loadMultiple();
        }
        else {
          $res = [$storage->load($match[2])];
        }

        foreach ($res as $entity) {
          if (!($entity instanceof ContentEntityInterface)) {
            $this->output()->writeln('Skipping non-content entity: ' . $entity->getEntityTypeId() . ':' . $entity->id());
            continue;
          }

          if (empty($entity)) {
            $this->output()->writeln('Entity not found: ' . $match[0]);
            continue;
          }

          $entities[$entity->getEntityTypeId() . ':' . $entity->id()] = $entity;
        }
      }
      catch (\Throwable $e) {
        $this->output()->writeln($e->getMessage());
      }
    }

    foreach ($entities as $entity) {
      if (!method_exists($entity, 'getFields')) {
        continue;
      }

      $entity_data = [
        'storage' => $entity->getEntityTypeId(),
      ];

      foreach ($entity->getFields() as $name => $field) {
        if (static::BANNED_FIELDS['any'][$name] ?? FALSE || static::BANNED_FIELDS[$entity->getEntityTypeId()][$name] ?? FALSE) {
          continue;
        }

        $entity_data['values'][$name] = $field->getValue();

        if (method_exists($field, 'referencedEntities')) {
          foreach ($field->referencedEntities() as $childEntity) {
            $childKey = $childEntity->getEntityTypeId() . ':' . $childEntity->id();

            if (!isset($entities[$childKey]) && !($childEntity instanceof ConfigEntityInterface)) {
              $references[$childKey] = $childKey;
            }
          }
        }
      }

      $data[] = $entity_data;
    }

    print(Yaml::encode($data));

    if (!empty($references)) {
      $this->output()->writeln("Unexported References: " . implode(' ', array_values($references)));
    }
  }

  /**
   * Export entity as configuration.
   *
   * @usage entity_yml:import < file.yml
   *  Imports yml and attempts to recreate entities defined within.
   *
   * @command entity_yml:import
   * @aliases eyi
   */
  public function entityImport() {
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $data = stream_get_contents(STDIN);
    $data = Yaml::decode($data);

    foreach ($data as $definition) {
      if (isset($definition['storage']) && is_string($definition['storage']) && isset($definition['values']) && is_array($definition['values'])) {
        try {
          $entityTypeManager->getStorage($definition['storage'])->create($definition['values'])->save();
        }
        catch (\Throwable $e) {
          $this->output()->writeln("*******************************************************************************");
          $this->output()->writeln("Error importing: " . print_r($definition, TRUE) . "\n" . $e->getMessage());
          $this->output()->writeln("*******************************************************************************");
          exit(1);
        }
      }
    }
  }

  /**
   * List entity types eligible for export.
   *
   * @usage entity_yml:list
   *  List entities eligible for export.
   *
   * @command entity_yml:list
   * @aliases eyl
   */
  public function entityList() {
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $entity_definitions = $entityTypeManager->getDefinitions();

    $list = [];

    foreach ($entity_definitions as $definition) {
      if ($definition instanceof ContentEntityTypeInterface) {
        $list[] = $definition->id();
      }
    }

    if (empty($list)) {
      $this->output()->writeln('No entity types eligible for export.');
    }
    else {
      $this->output()->writeln('Entity types eligible for export:');

      sort($list);

      foreach ($list as $entity_type) {
        $this->output()->writeln("\t" . $entity_type);
      }
    }
  }

}
